;***********************************************************************
;
;  6th order staggered derivatives, 5th order half zone staggering
;
;  15.6/92   kg: change dif{xyz}2 to halfpoint up
;
;***********************************************************************
FUNCTION xup, f 
;
;  f is centered on (i-.5,j,k), fifth order stagger
;
;
;-----------------------------------------------------------------------
      c = 3./256.
      b = -25./256.
      a = .5-b-c
      sx = size(f)
      if sx(0) eq 1 then begin
      return, a*(f + shift(f,-1)) $
          + b*(shift(f,1) + shift(f,-2)) $
          + c*(shift(f,2) + shift(f,-3))
      endif
      if sx(0) eq 2 then begin
      return, a*(f + shift(f,-1)) $
          + b*(shift(f,1,0) + shift(f,-2,0)) $
          + c*(shift(f,2,0) + shift(f,-3,0))
      endif
      return, a*(f + shift(f,-1,0,0)) $
          + b*(shift(f,1,0,0) + shift(f,-2,0,0)) $
          + c*(shift(f,2,0,0) + shift(f,-3,0,0))
END
;***********************************************************************
FUNCTION yup, f 
;
;  f is centered on (i,j-.5,k)
;
;-----------------------------------------------------------------------
      c = 3./256.
      b = -25./256.
      a = .5-b-c
      sx=size(f)
      if sx(0) eq 2 then begin
      return, a*(f + shift(f,0,-1)) $
          + b*(shift(f,0,1) + shift(f,0,-2)) $
          + c*(shift(f,0,2) + shift(f,0,-3))
      endif
      return, a*(f + shift(f,0,-1,0)) $
          + b*(shift(f,0,1,0) + shift(f,0,-2,0)) $
          + c*(shift(f,0,2,0) + shift(f,0,-3,0))
END
;***********************************************************************
FUNCTION zup, f 
;
;  f is centered on (i,j,k-.5)
;
;-----------------------------------------------------------------------
      c = 3./256.
      b = -25./256.
      a = .5-b-c
      return, a*(f + shift(f,0,0,-1)) $
          + b*(shift(f,0,0,1) + shift(f,0,0,-2)) $
          + c*(shift(f,0,0,2) + shift(f,0,0,-3))
END
;***********************************************************************
FUNCTION xdn, f 
;
;  f is centered on (i-.5,j,k)
;
;-----------------------------------------------------------------------
      c = 3./256.
      b = -25./256.
      a = .5-b-c
      sx = size(f)
      if sx(0) eq 1 then begin
      return, a*(f + shift(f,1)) $
          + b*(shift(f,-1) + shift(f,2)) $
          + c*(shift(f,-2) + shift(f,3))
      endif
      if sx(0) eq 2 then begin
      return, a*(f + shift(f,1,0)) $
          + b*(shift(f,-1,0) + shift(f,2,0)) $
          + c*(shift(f,-2,0) + shift(f,3,0))
      endif
      return, a*(f + shift(f,1,0,0)) $
          + b*(shift(f,-1,0,0) + shift(f,2,0,0)) $
          + c*(shift(f,-2,0,0) + shift(f,3,0,0))
END
;***********************************************************************
FUNCTION ydn, f 
;
;  f is centered on (i,j,k-.5)
;
;-----------------------------------------------------------------------
      c = 3./256.
      b = -25./256.
      a = .5-b-c
      sx = size(f)
      if sx(0) eq 2 then begin
      return, a*(f + shift(f,0,1)) $
          + b*(shift(f,0,-1) + shift(f,0,2)) $
          + c*(shift(f,0,-2) + shift(f,0,3))
      endif
      return, a*(f + shift(f,0,1,0)) $
          + b*(shift(f,0,-1,0) + shift(f,0,2,0)) $
          + c*(shift(f,0,-2,0) + shift(f,0,3,0))
END
;***********************************************************************
FUNCTION zdn, f 
;
;  f is centered on (i,j,k-.5)
;
;-----------------------------------------------------------------------
      c = 3./256.
      b = -25./256.
      a = .5-b-c
      return, a*(f + shift(f,0,0,1)) $
          + b*(shift(f,0,0,-1) + shift(f,0,0,2)) $
          + c*(shift(f,0,0,-2) + shift(f,0,0,3))
END
;***********************************************************************
FUNCTION ddxup, f 
;
;  X partial derivative
;
;-----------------------------------------------------------------------
      common grid, dx,dy,dz
      dx=1.
      c = (-1.+(3.^5-3.)/(3.^3-3.))/(5.^5-5.-5.*(3.^5-3))
      b = (-1.-120.*c)/24.
      a = (1.-3.*b-5.*c)/dx
      b = b/dx
      c = c/dx
      sx = size(f)
      if sx(0) eq 1 then begin
      return, a *(shift(f,-1) - f) $
          + b *(shift(f,-2) - shift(f,1)) $
          + c *(shift(f,-3) - shift(f,2))
      endif
      if sx(0) eq 2 then begin
      return, a *(shift(f,-1,0) - f) $
          + b *(shift(f,-2,0) - shift(f,1,0)) $
          + c *(shift(f,-3,0) - shift(f,2,0))
      endif
      return, a *(shift(f,-1,0,0) - f) $
          + b *(shift(f,-2,0,0) - shift(f,1,0,0)) $
          + c *(shift(f,-3,0,0) - shift(f,2,0,0))
END
;***********************************************************************
FUNCTION ddyup, f 
;
;  Y partial derivative
;
;-----------------------------------------------------------------------
      common grid, dx,dy,dz
      dy=1.
      c = (-1.+(3.^5-3.)/(3.^3-3.))/(5.^5-5.-5.*(3.^5-3))
      b = (-1.-120.*c)/24.
      a = (1.-3.*b-5.*c)/dy
      b = b/dy
      c = c/dy
      sx = size(f)
      if sx(0) eq 2 then begin
      return, a *(shift(f,0,-1) - f) $
          + b *(shift(f,0,-2) - shift(f,0,1)) $
          + c *(shift(f,0,-3) - shift(f,0,2))
      endif
      return, a *(shift(f,0,-1,0) - f) $
          + b *(shift(f,0,-2,0) - shift(f,0,1,0)) $
          + c *(shift(f,0,-3,0) - shift(f,0,2,0))
END
;***********************************************************************
FUNCTION ddzup, f 
;
;  Z partial derivative
;
;-----------------------------------------------------------------------
      common grid, dx,dy,dz
      dz=1.
      c = (-1.+(3.^5-3.)/(3.^3-3.))/(5.^5-5.-5.*(3.^5-3))
      b = (-1.-120.*c)/24.
      a = (1.-3.*b-5.*c)/dz
      b = b/dz
      c = c/dz
      return, a *(shift(f,0,0,-1) - f) $
          + b *(shift(f,0,0,-2) - shift(f,0,0,1)) $
          + c *(shift(f,0,0,-3) - shift(f,0,0,2))
END
;***********************************************************************
FUNCTION ddxdn, f 
;
;  X partial derivative
;
;-----------------------------------------------------------------------
      common grid, dx,dy,dz
      dx=1.
      c = (-1.+(3.^5-3.)/(3.^3-3.))/(5.^5-5.-5.*(3.^5-3))
      b = (-1.-120.*c)/24.
      a = (1.-3.*b-5.*c)/dx
      b = b/dx
      c = c/dx
      sx = size(f)
      if sx(0) eq 1 then begin
      return, a*(f - shift(f,1)) $
          + b*(shift(f,-1) - shift(f,2)) $
          + c*(shift(f,-2) - shift(f,3))
      endif
      if sx(0) eq 2 then begin
      return, a*(f - shift(f,1,0)) $
          + b*(shift(f,-1,0) - shift(f,2,0)) $
          + c*(shift(f,-2,0) - shift(f,3,0))
      endif
      return, a*(f - shift(f,1,0,0)) $
          + b*(shift(f,-1,0,0) - shift(f,2,0,0)) $
          + c*(shift(f,-2,0,0) - shift(f,3,0,0))
END
;***********************************************************************
FUNCTION ddydn, f 
;
;  Y partial derivative
;
;-----------------------------------------------------------------------
      common grid, dx,dy,dz
      dy=1.
      c = (-1.+(3.^5-3.)/(3.^3-3.))/(5.^5-5.-5.*(3.^5-3))
      b = (-1.-120.*c)/24.
      a = (1.-3.*b-5.*c)/dy
      b = b/dy
      c = c/dy
      sx = size(f)
      if sx(0) eq 2 then begin
      return, a*(f - shift(f,0,1)) $
          + b*(shift(f,0,-1) - shift(f,0,2)) $
          + c*(shift(f,0,-2) - shift(f,0,3))
      endif
      return, a*(f - shift(f,0,1,0)) $
          + b*(shift(f,0,-1,0) - shift(f,0,2,0)) $
          + c*(shift(f,0,-2,0) - shift(f,0,3,0))
END
;***********************************************************************
FUNCTION ddzdn, f 
;
;  Z partial derivative
;
;-----------------------------------------------------------------------
      common grid, dx,dy,dz
      dz=1.
      c = (-1.+(3.^5-3.)/(3.^3-3.))/(5.^5-5.-5.*(3.^5-3))
      b = (-1.-120.*c)/24.
      a = (1.-3.*b-5.*c)/dz
      b = b/dz
      c = c/dz
      return, a*(f - shift(f,0,0,1)) $
          + b*(shift(f,0,0,-1) - shift(f,0,0,2)) $
          + c*(shift(f,0,0,-2) - shift(f,0,0,3))
END
;-----------------------------------------------------------------------
