function vec_field,matrix,r_vec,y_period=y_period,z_period=z_period
;
; calculate function values of a field determiend by the matrix 'matrix'
; in the point r_vec(*,3) -> vecfield(*,3)
; done by linear interpolation in the matrix grid
;
; done by Klaus Galsgaard
;
;on_error,2             ;Return to caller if an error occurs
s=size(matrix)
sp=size(r_vec)

; field strength in points
vecfield=fltarr(sp(1),3,/nozero)

; matrix index for points
ix=long(r_vec(*,0))
iy=long(r_vec(*,1))
iz=long(r_vec(*,2))

; find points that are out side the box
ixu = where (ix lt 0, o1)
if o1 ne 0 then ix(ixu) = 0
ixo = where (ix gt s(1)-1, o2)          ; changed ge to gt
if o2 ne 0 then ix(ixo) = s(1)-2        ; changed s(1)-2 to s(1)-1
if keyword_set(y_period) eq 0 then begin
  iyu = where (iy le 0, o3)
  if o3 ne 0 then iy(iyu) = 0
  iyo = where (iy gt s(2)-1, o4)
  if o4 ne 0 then iy(iyo) = s(2)-2
endif
if keyword_set(z_period) eq 0 then begin
  izu = where (iz le 0, o5)
  if o5 ne 0 then iz(izu) = 0
  izo = where (iz gt s(3)-1, o6)
  if o6 ne 0 then iz(izo) = s(3)-2
endif

; relative position in the grid
p=r_vec(*,0)-ix
q=r_vec(*,1)-iy
r=r_vec(*,2)-iz

;print ,'max,min(iy)',max(iy,min=mi1),mi1, max(r_vec(*,1),min=mi),mi
; linear interpolation in the field grid
for k=0,2 do begin
  if keyword_set(y_period) or keyword_set(z_period) then begin
    zindex = ((iz+1) mod (s(3)))
    in = where (zindex lt 0, no)
    if (no ne 0) then zindex(in) = zindex(in) + s(3)
    yindex = ((iy+1) mod (s(2)))
    in = where (yindex lt 0, no)
    if (no ne 0) then yindex(in) = yindex(in) + s(2)
;    id1 = ((iz+1) mod (s(3)))*s(1)*s(2) + ((iy+1) mod (s(2)))*s(1) + ix +k*s(1)*s(2)*s(3)
    id1 = zindex*s(1)*s(2) + yindex*s(1) + ix +k*s(1)*s(2)*s(3)
    id2 = id1 + 1
;    id3 = ((iz+1) mod (s(3)))*s(1)*s(2) + iy*s(1) + ix + k*s(1)*s(2)*s(3)
    id3 = zindex*s(1)*s(2) + iy*s(1) + ix + k*s(1)*s(2)*s(3)
    id4 = id3 +1
;    id5 = (iz mod (s(3)))*s(1)*s(2) + ((iy+1) mod (s(2)))*s(1) + ix + k*s(1)*s(2)*s(3)
    id5 = zindex*s(1)*s(2) + yindex*s(1) + ix + k*s(1)*s(2)*s(3)
    id6 = id5 + 1
;    id7 = (iz mod (s(3)))*s(1)*s(2) + iy*s(1) + ix + k*s(1)*s(2)*s(3)
    id7 = zindex*s(1)*s(2) + iy*s(1) + ix + k*s(1)*s(2)*s(3)
    id8 = id7 + 1
  endif else begin
    id1 = (iz+1)*s(1)*s(2) + (iy+1)*s(1) + ix +k*s(1)*s(2)*s(3)
    id2 = id1 + 1
    id3 = (iz+1)*s(1)*s(2) + iy*s(1) + ix + k*s(1)*s(2)*s(3)
    id4 = id3 +1
    id5 = iz*s(1)*s(2) + (iy+1)*s(1) + ix + k*s(1)*s(2)*s(3)
    id6 = id5 + 1
    id7 = iz*s(1)*s(2) + iy*s(1) + ix + k*s(1)*s(2)*s(3)
    id8 = id7 + 1
  endelse
  vecfield(*,k)=r*( q* ( (1-p)*matrix(id1) +$
                             p*matrix(id2) ) +$
              (1-q)*(    (1-p)*matrix(id3)  +$
                             p*matrix(id4) ) ) +$
            (1-r)*( q* ( (1-p)*matrix(id5)   +$
                             p*matrix(id6) )   +$
                 (1-q)*( (1-p)*matrix(id7)    +$
                             p*matrix(id8) ) )
end

; only vecfield to the points that are inside the grid
; the points outside are put to zero
if o1 ne 0 then vecfield(ixu,*) = replicate(0,o1,3)
if o2 ne 0 then vecfield(ixo,*) = replicate(0,o2,3)
if keyword_set(y_period) eq 0 then begin
  if o3 ne 0 then vecfield(iyu,*) = replicate(0,o3,3)
  if o4 ne 0 then vecfield(iyo,*) = replicate(0,o4,3)
endif
if keyword_set(z_period) eq 0 then begin
  if o5 ne 0 then vecfield(izu,*) = replicate(0,o5,3)
  if o6 ne 0 then vecfield(izo,*) = replicate(0,o6,3)
endif

return,vecfield
end
