! $Id: fft.f90,v 1.1 2010/06/01 10:52:38 kg Exp $
!----------------------------------------------------------------------
!
!  Three-dimensional Fourier transform routines, based on FFTPACK.
!  FFTPACK is transportable, and allows arbitrary dimensions (speed
!  will vary, of course).
!
!  CALL fft3df (f, ft, mx, my, mz)              ! forward transform
!  CALL fft3db (ft, f, mx, my, mz)              ! backward transform
!  CALL fft3d_k2 (k2, dx, dy, dz, mx, my, mz)   ! compute k2
!
!**********************************************************************
SUBROUTINE fftpack_test
!
!  This test demonstrates that the coeffs returned are:
!
!   [a_0, +a_1/2., +b_1/2., ..., b_{ny-1}/2.,  a_ny]
!
!  Thus, the derivive is formed by setting the tranform to
!
!   [ 0., -b_1/2, +a_1/2., ...., a_{ny-1}/2.,  0.  ]
!
  real, allocatable, dimension(:):: f,ft,w
  integer m,lw,i,k
  real a,pi
!----------------------------------------------------------------------
  m=8
  pi=4.*atan(1.)
  lw=2*m+15
  allocate (f(m),ft(m),w(lw))

  a = 2.*pi/m
  f = 0.
  do k=0,m/2
    do i=1,m
      f(i) = f(i) + 2.*(k+1)*cos((i-1)*k*a) + 2.*(k+m)*sin((i-1)*k*a)
    end do
  end do
  call srffti (m,w)
  print '(a,8f8.4)','fft_test  f:', f
  ft = f
  call srfftf (m,ft,w)
  print '(a,8f8.4)','fft_test ft:', ft/m

  m = 7
  a = 2.*pi/m
  f = 0.
  do k=0,m/2
    do i=1,m
      f(i) = f(i) + 2.*(k+1)*cos((i-1)*k*a) + 2.*(k+m)*sin((i-1)*k*a)
    end do
  end do
  call srffti (m,w)
  print '(a,8f8.4)','fft_test  f:', f
  ft = f
  call srfftf (m,ft,w)
  print '(a,8f8.4)','fft_test ft:', ft

  deallocate (f,ft,w)
END

!**********************************************************************
SUBROUTINE fft_k2 (k2,dx1,dy1,dz1,mx1,my1,mz1)
  USE grid_m
  implicit none
  real, dimension(mx,my,mz):: k2
  real dx1, dy1, dz1, kx, ky, kz
  integer i, j, jj, k, kk, mx1, my1, mz1
!----------------------------------------------------------------------
  do k=1,mz
    kk=k+mpi_z*mz
    kz=2.*pi/sz*(kk/2)
    do j=1,my
      jj=j+mpi_y*my
      ky=2.*pi/sy*(jj/2)
      do i=1,mx
        kx=2.*pi/sx*(i/2)
        k2(i,j,k)=kx**2+ky**2+kz**2
      end do
    end do
  end do
END SUBROUTINE

!**********************************************************************
SUBROUTINE fft_f (r,rr,mx1,my1,mz1)
  USE grid_m
  implicit none
  integer i,j,k,mx1,my1,mz1
  integer, parameter:: lw=1024, lwx=2*lw+15
  real, dimension(mx,my,mz):: r, rr
  real c, wx(lwx), fx(mx), fy(my), fz(mz,mx)
!----------------------------------------------------------------------
  call fft_test
  if (mx.gt.lw .or. my.gt.lw .or. mz.gt.lw) then
    print *,'fft_f: dim too small',mx,my,mz,lw
    stop
  end if
  call srffti (mx,wx)

  do k=1,mz
   do j=1,my
    fx(:)=r(:,j,k)
    call srfftf (mx,fx,wx)
    c=1./mx
    scratch(:,j,k)=c*fx(:)
   end do
  end do

  call permute (scratch, rr)
  do k=1,mz
   do j=1,my
    fx(:)=rr(:,j,k)
    call srfftf (mx,fx,wx)
    c=1./mx
    scratch(:,j,k)=c*fx(:)
   end do
  end do

  call permute (scratch, rr)
  do k=1,mz
   do j=1,my
    fx(:)=rr(:,j,k)
    call srfftf (mx,fx,wx)
    c=1./mx
    scratch(:,j,k)=c*fx(:)
   end do
  end do

  call permute (scratch, rr)

END SUBROUTINE

!**********************************************************************
SUBROUTINE fft_r (rr,r,mx1,my1,mz1)
  USE grid_m
  implicit none
  integer i,j,k,mx1,my1,mz1
  integer, parameter:: lw=1024, lwx=2*lw+15

  real, dimension(mx,my,mz):: r, rr
  real wx(lwx), fx(mx), fy(my), fz(mz,mx)
!----------------------------------------------------------------------

  if (mx.gt.lw .or. my.gt.lw .or. mz.gt.lw) then
    print *,'fft_r: dim too small',mx,my,mz,lw
    stop
  end if
  call srffti (mx,wx)

  call permute (rr, scratch)
  do k=1,mz
   do j=1,my
    fx=scratch(:,j,k)
    call srfftb (mx,fx,wx)
    r(:,j,k)=fx
   end do
  end do

  call permute (r, scratch)
  do k=1,mz
   do j=1,my
    fx=scratch(:,j,k)
    call srfftb (mx,fx,wx)
    r(:,j,k)=fx
   end do
  end do

  call permute (r, scratch)
  do k=1,mz
   do j=1,my
    fx=scratch(:,j,k)
    call srfftb (mx,fx,wx)
    r(:,j,k)=fx
   end do
  end do

END SUBROUTINE

!**********************************************************************
SUBROUTINE fft_test
  USE grid_m
  implicit none
  integer ix, iy, iz, seed2
  real random
  logical, save:: first_time=.true.
  real, allocatable, dimension(:,:,:):: fft1, fft2, fft3
!----------------------------------------------------------------------
  if (.not. first_time) return
  first_time = .false.

  if (mpi_mx > 1) then
    if (master) print *,'WARNING, fft3d_mpi2.f90: cannot handle MPI in x-direction; skipping tests'
    return
  end if

  if (master) call fftpack_test
  call test_permute

  allocate (fft1(mx,my,mz), fft2(mx,my,mz), fft3(mx,my,mz))
  seed2 = 0
  do iz=1,mz
  do iy=1,my
  do ix=1,mx
    !fft1(ix,iy,iz) = cos(2.*pi*(x(ix)-xmin)/sx) + 2.*cos(2.*pi*(y(iy)-ymin)/sy) + 3.*cos(2.*pi*(z(iz)-zmin)/sz)
    fft1(ix,iy,iz) = random(seed2)
  end do
  end do
  end do

  call fft_f (fft1, fft2, mx, my, mz)
  if (master) print '(a,8f10.3)', 'fft_test:', fft2(1:2,1:2,1:2)

  call fft_r (fft2, fft3, mx, my, mz)
  if (maxval(abs(fft1-fft3)) > 1e-6) print*,mpi_rank,': fft_test WARNING:',maxval(abs(fft1-fft3))

  deallocate (fft1, fft2, fft3)
END
