! $Id: support.f90,v 1.1 2010/06/01 10:54:01 kg Exp $
!*******************************************************************************
SUBROUTINE init_mpi
END

!*******************************************************************************
SUBROUTINE end_mpi
END

!*******************************************************************************
SUBROUTINE init_mpi_grid
  USE grid_m
  implicit none
!...............................................................................
  xmin_mpi = xmin; ymin_mpi=ymin; zmin_mpi=zmin
  xmax_mpi = xmax; ymax_mpi=ymax; zmax_mpi=zmax
END

!*******************************************************************************
REAL FUNCTION fmaxval (f)
  USE grid_m
  implicit none
!...............................................................................
  real f(mx,my,mz)
  fmaxval = maxval(f)
END FUNCTION

!*******************************************************************************
REAL FUNCTION fminval (f)
  USE grid_m
  implicit none
!...............................................................................
  real f(mx,my,mz)
  fminval = minval(f)
END FUNCTION

!*******************************************************************************
REAL FUNCTION faver (f)
  USE grid_m
  implicit none
  real f(mx,my,mz)
  real(kind=8):: ysum, zsum
  integer iy, iz
!...............................................................................
  zsum = 0.
  do iz=1,mz
    ysum = 0.
    do iy=1,my
      ysum = ysum+sum(f(:,iy,iz))
    end do
    zsum = zsum+ysum
  end do
  faver = zsum/(real(mx)*real(my)*real(mz))
END FUNCTION

!*******************************************************************************
REAL FUNCTION fsum (f)
  implicit none
  real f
  fsum = f
END FUNCTION

!*******************************************************************************
SUBROUTINE write_snapshot 
  USE io_m
  USE grid_m
  implicit none
  character(len=5) nrank
  namelist /grid/ mx, my, mz, sx, sy, sz, mpi_mx, mpi_my, mpi_mz
!.......................................................................
  if (t < tnext) return                                                         ! snap shot?
  if (master) print'(1x,a,i4,a,f9.6,3x,a)','write_snapshot: ',iwrite, &
    ' time =',t,'snapshot.dat'

  open (data_unit,file='grid.nml', &
        status='unknown',form='formatted')                                      ! write .nml file
  write (data_unit,grid)                                                        ! namelist -> grid size
  close (data_unit)

  open (data_unit,file='time.dat', access='direct',recl=4*3)                    ! time data file    
  write (data_unit,rec=1+(iwrite-1)) t,dt                                  ! to enable restarts
  close (data_unit)
  open (data_unit,file='snapshot.dat', &
        access='direct',recl=4*mx*my*mz)                                        ! write snapshot file
  write (data_unit,rec=1+(iwrite-1)*mvar) rho                                   ! read density ..
  write (data_unit,rec=2+(iwrite-1)*mvar) px                                    ! .. momentum ..
  write (data_unit,rec=3+(iwrite-1)*mvar) py                                    ! ..
  write (data_unit,rec=4+(iwrite-1)*mvar) pz                                    ! ..
  write (data_unit,rec=5+(iwrite-1)*mvar) e                                     ! ..
  write (data_unit,rec=6+(iwrite-1)*mvar) Bx                                    ! .. magnetic field ..
  write (data_unit,rec=7+(iwrite-1)*mvar) By                                    ! ..
  write (data_unit,rec=8+(iwrite-1)*mvar) Bz                                    ! ..
  close (data_unit)

  tnext = tnext + tsnap                                                         ! next snapshot time
  iwrite = iwrite + 1                                                           ! increment snap shot number
END SUBROUTINE

!*******************************************************************************
SUBROUTINE read_snapshot  
  USE io_m
  USE grid_m
  implicit none
  integer i
  namelist /grid/ mx, my, mz, sx, sy, sz
!...............................................................................
  open (data_unit,file='time.dat',access='direct',recl=4*3)                     ! time data file
  read (data_unit,rec=iread) t,dt                                          ! ..
  close (data_unit)                                                             ! close the file
  if (master) print'(1x,a,i4,a,f9.6,3x,a)','read_snapshot: ',iread, &
    ' time =',t,'snapshot.dat'

  open (data_unit,file='snapshot.dat',form='unformatted', &
        status='old',access='direct',recl=4*mx*my*mz)                           ! snapshot data file
  read (data_unit,rec=1+(iread-1)*mvar) rho                                     ! read density ..
  read (data_unit,rec=2+(iread-1)*mvar) px                                      ! .. momentum ..
  read (data_unit,rec=3+(iread-1)*mvar) py                                      ! ..
  read (data_unit,rec=4+(iread-1)*mvar) pz                                      ! ..
  read (data_unit,rec=5+(iread-1)*mvar) e                                       ! .. thermal energy
  read (data_unit,rec=6+(iread-1)*mvar) Bx                                      ! .. magnetic field ..
  read (data_unit,rec=7+(iread-1)*mvar) By                                      ! ..
  read (data_unit,rec=8+(iread-1)*mvar) Bz                                      ! ..
  close (data_unit)                                                             ! close the file
  tnext = (ifix(t/tsnap)+1)*tsnap                                               ! next snapshot time
END SUBROUTINE
