! $Id: io.f90,v 1.1 2010/06/01 10:54:58 kg Exp $
!*******************************************************************************
MODULE io_m  
  USE grid_m
  implicit none
  real tsnap                                                                    ! time between snapshots
  real tnext                                                                    ! next snapshot time
  integer nprint, it                                                            ! snapshot number, total steps
  integer nsnap                                                                 ! number of snapshots, and index
  integer iread, iwrite                                                         ! read and write snapshot numbers
END MODULE

!*******************************************************************************
SUBROUTINE init_io 
  USE io_m
  implicit none
  namelist /io/ iread, iwrite, tsnap, nsnap                                     ! output control namelist
!...............................................................................
  tsnap = 0.1
  nsnap = 10
  tend  = 1.0
  iread = -1                                                                    ! initial data
  iwrite = -1                                                                   ! first snapashot
  rewind(txt_input); read (txt_input,io)                                        ! output control namelist
  if (iwrite == -1) iwrite = iread+1                                            ! default output snapshot
  if (master) write (*,io)                                                      ! write to log
  tnext = t                                                                     ! snapshot time
  tend  = t + nsnap*tsnap                                                       ! ending time
END SUBROUTINE

!*******************************************************************************
REAL FUNCTION wallclock()                                                       ! wall clock timer
  implicit none
  integer, save:: count, count_rate=0, count_max
  real, save:: previous=0., offset=0.
!...............................................................................
  if (count_rate == 0) then
    call system_clock(count=count, count_rate=count_rate, count_max=count_max)
    offset = -count/real(count_rate)                                            ! startup offset (seconds)
  else
    call system_clock(count=count)
  end if
  wallclock = count/real(count_rate) + offset                                   ! seconds
  if (wallclock < previous) then                                                ! handle wrap-around
    offset = offset + count_max/real(count_rate)                                ! add wrap-around offset
    wallclock = count/real(count_rate) + offset
  end if
  previous = wallclock
END
