! $Id: main.f90,v 1.1 2010/06/01 10:54:58 kg Exp $
!*******************************************************************************
PROGRAM Fourier_Field
!
!  Main program for the evolution of a simple Fourier defined field with internal
!  structure and forces that drives the system in to a collpase
!
!-------------------------------------------------------------------------------
  USE grid_m, only: mx, my, mz                                                  ! number of grid points
  USE io_m, only: t, dt, tend, it, txt_input, master                            ! input/output 
  implicit none                                                                 ! keep us straight
  real wc0, wc1, wallclock                                                      ! wall clock times
!...............................................................................
  open (txt_input,file='input.txt',status='old')                                ! namelist input unit         (input.txt)
  call init_mpi                                                                 ! initialize MPI              (mpi.f90)
  call init_grid                                                                ! initialize grid arrays      (grid.f90)
  call init_mhd                                                                 ! initialize MHD routine      (mhd.f90)
  call init_io                                                                  ! initialize input/output     (io.f90)
  call init_timestep                                                            ! initialize time stepping    (timestep.f90)
  call init_values                                                               ! initialize field setup     (fourier_field.f90)

  wc0 = wallclock()                                                             ! initial wall clock time     (io.f90)
  do while (t < tend+0.5*dt)                                                    ! loop (avoiding round-off)
    call timestep                                                               ! time step                   (timestep.f90)
    call write_snapshot                                                         ! time to write snapshot?     (io.f90)
  end do
  wc1 = wallclock()                                                             ! final wall clock time       (io.f90)

  call deallocate_arrays                                                        ! avoids g95 complaints       (grid.f90)
  if (master) print '(a,f5.2,a)','done:', (wc1-wc0)*1e6 &
    /(real(it)*real(mx)*real(my)*real(mz)),' CPU-musec/zone-update'             ! CPU-microsec per zone-update
       
  call end_mpi                                                                  ! finalize MPI                (mpi.f90)
END PROGRAM
